#include"phisq.h"
#include<string>
#include <iomanip>
PHISQ::PHISQ(int L,double beta,int nsim,int verb): L(L), beta(beta), nsim(nsim),verb(verb)
{

    Ntotal=L*L;
    cout<<"lattice LxL "<<L<<" "<<Ntotal<<" beta "<<beta<<std::endl;
    Phi.resize(Ntotal);
    Cluster.resize(Ntotal);
    Stack.reserve(Ntotal);
    IndCluster.reserve(Ntotal);
    // creer a nouveau un mersenne aleatoire et recopier le pointeur sur gen
    mt19937_64 GenerRd(rd());
    gen=GenerRd;
    //mt19937 gen(1000); Set the initial condition
    proba=1-exp(-2.0*beta);
    cout<<Ntotal<<std::endl;
    thermofichier.open("energie"+to_string(Ntotal)+"beta"+to_string(beta)+"sim"+to_string(nsim)+".dat");
    if (verb==1) clusterfichier.open("cluster"+to_string(Ntotal)+"beta"+to_string(beta)+"sim"+to_string(nsim)+".dat");
    Neigh= new int [Ntotal][4];
    for (int i=0; i<L; ++i)
    {
        for (int j=0; j<L; ++j)
        {
            Neigh[i*L+j][0]=((i+1)%L)*L+j;
            Neigh[i*L+j][1]=((i-1+L)%L)*L+j;
            Neigh[i*L+j][2]=i*L+(j+1)%L;
            Neigh[i*L+j][3]=i*L+(j-1+L)%L;
        }
    }
    Energy=energycheck();
    Magne=magnecheck();
}
void PHISQ::Tryadd(int i, int cluster_spin)
{
    uniform_real_distribution <double> disdu(0,1.0);
    if (!Cluster[i] && Phi[i]==cluster_spin)
        if (disdu(gen)< proba)
        {
            //modification de Ferrenberg Landau
            Phi[i]=0;
            IndCluster.push_back(i);
            Cluster[i]=1;
            Stack.push_back(i);
        }
}

void PHISQ::performHK(int eq) //Hoshen-Kopelman  stepper
{
    uniform_int_distribution <int> distint(0,Ntotal-1);
    Stack.resize(0);
    IndCluster.resize(0);
    for (auto &i :Cluster) {i=0;}
    // reset Cluster array
    int phitry=distint(gen); // choose a random site uniformly
    cluster_spin =Phi[phitry]; //signe du cluster

    //modification for implementing an efficient update energy
    Phi[phitry]=0;
    Cluster[phitry]=1;
    Stack.push_back(phitry);
    IndCluster.push_back(phitry);
    while (Stack.size()>0)
    {
        int itry=Stack.back();
        Stack.pop_back();
        //scan neighbors
        for (int j=0; j<4; j++) // TryAdd(Neigh[phitry,0],cluster_spin);
        {
            Tryadd(Neigh[itry][j],cluster_spin);
        }

    }
    for (auto i :IndCluster)
    {
        for (int j=0; j<4; ++j)
        {
            if (Phi[Neigh[i][j]] != Phi[i])
                Energy+=2*Phi[Neigh[i][j]]*cluster_spin;
        }
    }

    // update des spins
    for (auto i:IndCluster) Phi[i]=-cluster_spin;
    Magne-= 2.0*cluster_spin*IndCluster.size();
    if (eq>0 && verb==1)
    {
        clusterfichier<<IndCluster.size()<<endl;
        if(eq==2) clusterfichier.close();
    }
}
void PHISQ::cold(void)
{
    for(int i=0; i<Ntotal; ++i)
        Phi[i] = 1.0;
    Magne=Ntotal;
    Energy=-2.0*Ntotal;
    cout<<"magne "<<Magne/(double) Ntotal<<" energy "<<Energy<<" beta "<<beta<<endl;
}
void PHISQ::hot(void)
{
    uniform_int_distribution <int> distint2(0,1);
    //std::uniform_real_distribution <double> disd(-1.0,1.0);
    for(int i=0; i<Ntotal; ++i)
    {
        Phi[i] = -1+2*distint2(gen);
        Magne+=Phi[i];
    }
    for (int i =0; i<Ntotal; ++i)
    {
        Energy+=-(Phi[Neigh[i][0]]+Phi[Neigh[i][2]])*Phi[i];
    }
    cout<<"magne "<<Magne/(double)Ntotal<<" energy "<<Energy<<endl;

}
double PHISQ::energycheck()
{
    double   Energy2=0;
    for (int i =0; i<Ntotal; ++i)
    {
        Energy+=-(Phi[Neigh[i][0]]+Phi[Neigh[i][2]])*Phi[i];
    }
    return(Energy2);
}

double PHISQ::magnecheck()
{
    double Magne2=0;
    for (auto i : Phi)
    {
        Magne2+=i;
    }

    return(Magne2);
}

void PHISQ::thermo_update(int eq)
{

    if (eq>0)
    {
        thermofichier<<setprecision(10)<<Energy<<" "<<Magne<<endl;
        if(eq==2) thermofichier.close();
    }
}

